#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2020 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/master/LICENSE.md

"""
Tasks Module

File : tasks.py

Brief : This file contains the definitions and functionalities for interacting
        with Tasks for a given Redfish service
"""

import sys
import time

def poll_task_monitor( context, response ):
    """
    Monitors a task monitor until it's complete and prints out progress
    NOTE: This call will block until the task is complete

    Args:
        context: The Redfish client object with an open session
        response: The initial response from the operation that produced the task

    Returns:
        The final response of the request
    """

    # No task was produced; just return the existing response
    if not response.is_processing:
        return response

    # Poll the task until completion
    task_monitor = response
    while task_monitor.is_processing:
        # Print the progress
        task_state = None
        task_percent = None
        if task_monitor.dict is not None:
            task_state = task_monitor.dict.get( "TaskState", None )
            task_percent = task_monitor.dict.get( "PercentComplete", None )
        if task_state is None:
            task_state = "Running"
        if task_percent is None:
            progress_str = "Task is {}\r".format( task_state )
        else:
            progress_str = "Task is {}: {}% complete\r".format( task_state, task_percent )
        sys.stdout.write( "\x1b[2K" )
        sys.stdout.write( progress_str )
        sys.stdout.flush()

        # Sleep for the requested time
        retry_time = response.retry_after
        if retry_time is None:
            retry_time = 1
        time.sleep( retry_time )

        # Check the monitor for an update
        task_monitor = response.monitor( context )
    sys.stdout.write( "\x1b[2K" )
    print( "Task is Done!" )

    return task_monitor



#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2020 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/master/LICENSE.md

"""
Redfish Sensor List

File : rf_sensor_list.py

Brief : This script uses the redfish_utilities module to dump sensor info
"""

import argparse
import redfish
import redfish_utilities

# Get the input arguments
argget = argparse.ArgumentParser( description = "A tool to walk a Redfish service and list sensor info" )
argget.add_argument( "--user", "-u", type = str, required = True, help = "The user name for authentication" )
argget.add_argument( "--password", "-p",  type = str, required = True, help = "The password for authentication" )
argget.add_argument( "--rhost", "-r", type = str, required = True, help = "The address of the Redfish service (with scheme)" )
args = argget.parse_args()

# Set up the Redfish object
redfish_obj = redfish.redfish_client( base_url = args.rhost, username = args.user, password = args.password )
redfish_obj.login( auth = "session" )

try:
    # Get and print the sensor info
    sensors = redfish_utilities.get_sensors( redfish_obj )
    redfish_utilities.print_sensors( sensors )
finally:
    # Log out
    redfish_obj.logout()



#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2020 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/master/LICENSE.md

"""
Inventory Module

File : inventory.py

Brief : This file contains the definitions and functionalities for scanning a
        Redfish service for an inventory of components
"""

import xlsxwriter

def get_system_inventory( context ):
    """
    Walks a Redfish service for system component information, such as drives,
    processors, and memory

    Args:
        context: The Redfish client object with an open session

    Returns:
        A list containing all system component information
    """

    inventory_list = []

    # Get the Service Root to find the Chassis Collection
    service_root = context.get( "/redfish/v1/" )
    if "Chassis" not in service_root.dict:
        # No Chassis Collection
        return inventory_list

    # Get the Chassis Collection and iterate through its collection
    chassis_col = context.get( service_root.dict["Chassis"]["@odata.id"] )
    for chassis_member in chassis_col.dict["Members"]:
        chassis = context.get( chassis_member["@odata.id"] )

        # Catalog Chassis itself
        chassis_instance = {
            "ChassisName": chassis.dict["Id"],
            "Chassis": [],
            "Processors": [],
            "Memory": [],
            "Drives": [],
            "PCIeDevices": [],
            "StorageControllers": [],
            "NetworkAdapters": []
        }
        inventory_list.append( chassis_instance )
        catalog_resource( chassis.dict, chassis_instance["Chassis"] )

        # Catalog all Drives, PCIeDevices, NetworkAdapters, Systems, and ResourceBlocks in the Chassis
        if "Links" in chassis.dict:
            catalog_array( context, chassis.dict["Links"], "Drives", chassis_instance["Drives"] )
            catalog_array( context, chassis.dict["Links"], "PCIeDevices", chassis_instance["PCIeDevices"] )
            catalog_systems( context, chassis.dict["Links"], "ComputerSystems", chassis_instance )
        catalog_collection( context, chassis.dict, "NetworkAdapters", chassis_instance["NetworkAdapters"] )

    return inventory_list

def catalog_array( context, resource, name, inventory ):
    """
    Catalogs an array of resources for the inventory list

    Args:
        context: The Redfish client object with an open session
        resource: The resource with the array
        name: The name of the property of the array
        inventory: The inventory list to update
    """

    if name in resource:
        for member in resource[name]:
            member_res = context.get( member["@odata.id"] )
            catalog_resource( member_res.dict, inventory )

def catalog_collection( context, resource, name, inventory ):
    """
    Catalogs a collection of resources for the inventory list

    Args:
        context: The Redfish client object with an open session
        resource: The resource with the collection
        name: The name of the property of the collection
        inventory: The inventory list to update
    """

    if name in resource:
        collection = context.get( resource[name]["@odata.id"] )
        for member in collection.dict["Members"]:
            member_res = context.get( member["@odata.id"] )
            catalog_resource( member_res.dict, inventory )

def catalog_systems( context, resource, name, inventory ):
    """
    Catalogs an array of systems for the inventory list

    Args:
        context: The Redfish client object with an open session
        resource: The resource with the array of computer systems
        name: The name of the property of the array
        inventory: The inventory list to update
    """

    if name in resource:
        for system in resource[name]:
            system_res = context.get( system["@odata.id"] )

            # Catalog all Processors, Memory, and PCIeDevices in the System
            catalog_collection( context, system_res.dict, "Processors", inventory["Processors"] )
            catalog_collection( context, system_res.dict, "Memory", inventory["Memory"] )
            catalog_array( context, system_res.dict, "PCIeDevices", inventory["PCIeDevices"] )
            catalog_simple_storage( context, system_res.dict, "SimpleStorage", inventory["Drives"] )
            catalog_storage( context, system_res.dict, "Storage", inventory )

def catalog_simple_storage( context, resource, name, inventory ):
    """
    Catalogs a collection of Simple Storage resources for the inventory list

    Args:
        context: The Redfish client object with an open session
        resource: The resource with the collection
        name: The name of the property of the collection
        inventory: The inventory list to update
    """

    if name in resource:
        collection = context.get( resource[name]["@odata.id"] )
        for member in collection.dict["Members"]:
            member_res = context.get( member["@odata.id"] )
            if "Devices" in member_res.dict:
                for index, drive in enumerate( member_res.dict["Devices"] ):
                    drive["@odata.id"] = "{}#/Devices/{}".format( member_res.dict["@odata.id"], index )
                    drive["@odata.type"] = "#Drive.Drive"
                    drive["Id"] = drive["Name"]
                    catalog_resource( drive, inventory )

def catalog_storage( context, resource, name, inventory ):
    """
    Catalogs a collection of Storage resources for the inventory list

    Args:
        context: The Redfish client object with an open session
        resource: The resource with the collection
        name: The name of the property of the collection
        inventory: The inventory list to update
    """

    if name in resource:
        collection = context.get( resource[name]["@odata.id"] )
        for member in collection.dict["Members"]:
            member_res = context.get( member["@odata.id"] )
            catalog_array( context, member_res.dict, "Drives", inventory["Drives"] )
            if "StorageControllers" in member_res.dict:
                for index, controller in enumerate( member_res.dict["StorageControllers"] ):
                    controller["@odata.type"] = "#StorageController.StorageController"
                    controller["Id"] = controller["MemberId"]
                    catalog_resource( controller, inventory["StorageControllers"] )

def catalog_resource( resource, inventory ):
    """
    Catalogs a resource for the inventory list

    Args:
        resource: The resource to catalog
        inventory: The inventory list to update
    """

    # Scan the inventory to ensure this is a new entry
    for item in inventory:
        if item["Uri"] == resource["@odata.id"]:
            return

    resource_type = resource["@odata.type"].rsplit( "." )[-1]

    location_prop = "Location"
    if resource_type == "Drive":
        location_prop = "PhysicalLocation"

    catalog = {
        "Uri": resource["@odata.id"],
        "PartNumber": resource.get( "PartNumber", None ),
        "SerialNumber": resource.get( "SerialNumber", None ),
        "Manufacturer": resource.get( "Manufacturer", None ),
        "Model": resource.get( "Model", None ),
        "SKU": resource.get( "SKU", None ),
        "AssetTag": resource.get( "AssetTag", None ),
        "Label": resource.get( location_prop, {} ).get( "PartLocation", {} ).get( "ServiceLabel", None ),
        "State": resource.get( "Status", {} ).get( "State", None ),
        "Description": None
    }

    # If no label was found, build a default name
    if catalog["Label"] is None:
        catalog["Label"] = resource_type + ": " + resource["Id"]

    # Build a string description of the component based on other properties
    prop_list = []
    if resource_type == "Chassis":
        prop_list = [ "Model" ]
    elif resource_type == "Processor":
        if catalog["Model"] is not None:
            prop_list = [ "Model" ]
        else:
            prop_list = [ "Manufacturer", "ProcessorArchitecture", "ProcessorType", "TotalCores", "MaxSpeedMHz" ]
    elif resource_type == "Memory":
        prop_list = [ "Manufacturer", "CapacityMiB", "MemoryDeviceType", "MemoryType" ]
    elif resource_type == "Drive":
        prop_list = [ "Manufacturer", "CapacityBytes", "Protocol", "MediaType" ]
    elif resource_type == "PCIeDevice":
        prop_list = [ "Manufacturer", "Model", "DeviceType", "PCIeInterface" ]
    elif resource_type == "StorageController":
        prop_list = [ "Manufacturer", "SpeedGbps", "SupportedDeviceProtocols" ]
    elif resource_type == "NetworkAdapter":
        prop_list = [ "Manufacturer", "Model" ]

    # Based on the listed properties for the resource type, build the description string
    if catalog["State"] != "Absent":
        description_str = ""
        for prop in prop_list:
            if resource.get( prop, None ) is not None:
                prop_val = resource[prop]
                # Some properties require refinement
                if prop == "TotalCores":
                    prop_val = str( prop_val ) + " Cores"
                elif prop == "MaxSpeedMHz":
                    prop_val = "@ " + str( prop_val ) + "MHz"
                elif prop == "CapacityMiB":
                    prop_val = str( prop_val ) + "MB"
                elif prop == "CapacityBytes":
                    prop_val = str( int( prop_val / ( 2 ** 30 ) ) ) + "GB"
                elif prop == "SpeedGbps":
                    prop_val = str( prop_val ) + "Gbps"
                elif prop == "SupportedDeviceProtocols":
                    prop_val = "/".join( prop_val ) + " Controller"
                elif prop == "DeviceType":
                    prop_val = prop_val + " PCIe Device"
                elif prop == "PCIeInterface":
                    if "MaxPCIeType" in prop_val:
                        prop_val = "@" + " " + prop_val["MaxPCIeType"]
                    else:
                        continue
                description_str = description_str + " " + prop_val
            else:
                # Some properties will have a default if not found
                if prop == "MediaType":
                    description_str = description_str + " Drive"
                elif prop == "SupportedDeviceProtocols":
                    description_str = description_str + " Storage Controller"
        catalog["Description"] = description_str.strip()

    inventory.append( catalog )

def print_system_inventory( inventory_list, details = False, skip_absent = False ):
    """
    Prints the system inventory list into a table

    Args:
        inventory_list: The inventory list to print
        details: True to print all of the detailed info
        skip_absent: True to skip printing absent components
    """

    inventory_line_format = "  {:35s} | {}"
    inventory_line_format_detail = "  {:35s} | {}: {}"
    inventory_line_format_empty = "  {:35s} | Not Present"

    # Go through each chassis instance
    for chassis in inventory_list:
        print( "'" + chassis["ChassisName"] + "' Inventory" )
        print( inventory_line_format.format( "Name", "Description" ) )

        # Go through each component type in the chassis
        type_list = [ "Chassis", "Processors", "Memory", "Drives", "PCIeDevices", "StorageControllers", "NetworkAdapters" ]
        for inv_type in type_list:
            # Go through each component and prints its info
            for item in chassis[inv_type]:
                if item["State"] == "Absent":
                    if not skip_absent:
                        print( inventory_line_format_empty.format( item["Label"][:35] ) )
                else:
                    print( inventory_line_format.format( item["Label"][:35], item["Description"] ) )

                    if details:
                        detail_list = [ "Manufacturer", "Model", "SKU", "PartNumber", "SerialNumber", "AssetTag" ]
                        for detail in detail_list:
                            if item[detail] is not None:
                                print( inventory_line_format_detail.format( "", detail, item[detail] ) )
        print( "" )


def write_system_inventory( inventory_list, file_name ):
    """
    Write the system inventory list into a spreadsheet

    Args:
        inventory_list: The inventory list to write to an Excel spreadsheet
        file_name: The name of the file for the spreadsheet
    """

    # Excel workbook to save data extracted and parsed
    workbook = xlsxwriter.Workbook( "./{}.xlsx".format( file_name ) )

    worksheet = workbook.add_worksheet( "Device Inventory" )
    cell_header_format = workbook.add_format( { 'bold': True, 'bg_color': 'yellow' } )
    cell_name_format = workbook.add_format( { 'bold': True } )

    column = 0
    row = 0

    # Adds header to Excel file
    header = [ "NAME", "DESCRIPTION", "MANUFACTURER", "MODEL", "SKU", "PART NUMBER", "SERIAL NUMBER", "ASSET TAG" ]
    for column_title in header:
        worksheet.write( row, column, column_title, cell_header_format )
        column += 1
    row = 1

    for chassis in inventory_list:
        # Go through each component type in the chassis
        type_list = [ "Chassis", "Processors", "Memory", "Drives", "PCIeDevices", "StorageControllers", "NetworkAdapters" ]
        for inv_type in type_list:
            # Go through each component and prints its info
            for item in chassis[inv_type]:
                column = 0
                worksheet.write( row, column, inv_type, cell_name_format )
                column += 1
                detail_list = [ "Description", "Manufacturer", "Model", "SKU", "PartNumber", "SerialNumber", "AssetTag" ]
                for detail in detail_list:
                    worksheet.write( row, column, item[detail] )
                    column += 1
                row += 1

    workbook.close()



#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2020 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/master/LICENSE.md

"""
Redfish BIOS Settings

File : rf_bios_settings.py

Brief : This script uses the redfish_utilities module to manager BIOS settings of a system
"""

import argparse
import redfish
import redfish_utilities

# Get the input arguments
argget = argparse.ArgumentParser( description = "A tool to manager BIOS settings for a system" )
argget.add_argument( "--user", "-u", type = str, required = True, help = "The user name for authentication" )
argget.add_argument( "--password", "-p",  type = str, required = True, help = "The password for authentication" )
argget.add_argument( "--rhost", "-r", type = str, required = True, help = "The address of the Redfish service (with scheme)" )
argget.add_argument( "--system", "-s", type = str, help = "The ID of the system to manage" )
argget.add_argument( "--attribute", "-a", type = str, nargs = 2, metavar = ( "name", "value" ), action = "append", help = "Sets a BIOS attribute to a new value; can be supplied multiple times to set multiple attributes" )
args = argget.parse_args()

# Set up the Redfish object
redfish_obj = redfish.redfish_client( base_url = args.rhost, username = args.user, password = args.password )
redfish_obj.login( auth = "session" )

try:
    # Get the BIOS settings
    current_settings, future_settings = redfish_utilities.get_system_bios( redfish_obj, args.system )

    if args.attribute is not None:
        new_settings = {}
        for attribute in args.attribute:
            # Based on the current settings, determine the appropriate data type for the new setting
            new_value = attribute[1]
            if attribute[0] in current_settings:
                if isinstance( current_settings[attribute[0]], bool ):
                    # Boolean; convert from a string
                    if new_value.lower() == "true":
                        new_value = True
                    else:
                        new_value = False
                elif isinstance( current_settings[attribute[0]], ( int, float ) ):
                    # Integer or float; go by the user input to determine how to convert since the current value may be truncated
                    try:
                        new_value = int( new_value )
                    except:
                        new_value = float( new_value )

            # Set the specified attribute to the new value
            new_settings[attribute[0]] = new_value
            print( "Setting {} to {}...".format( attribute[0], attribute[1] ) )
        redfish_utilities.set_system_bios( redfish_obj, new_settings, args.system )
    else:
        # Print the BIOS settings
        redfish_utilities.print_system_bios( current_settings, future_settings )
finally:
    # Log out
    redfish_obj.logout()



#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2020 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/master/LICENSE.md

"""
Logs Module

File : logs.py

Brief : This file contains the definitions and functionalities for interacting
        with the log service for a given Redfish service
"""

import shutil

from .messages import verify_response
from enum import Enum

class RedfishLogServiceNotFoundError( Exception ):
    """
    Raised when a matching log service cannot be found
    """
    pass

class RedfishClearLogNotFoundError( Exception ):
    """
    Raised when a log service does not contain the clear log action
    """
    pass

class log_container( Enum ) :
    """
    Types of resources that contain log services
    """
    MANAGER = "Managers"
    CHASSIS = "Chassis"
    SYSTEM = "Systems"

def get_log_entries( context, container_type = log_container.MANAGER, container_id = None, log_service_id = None ):
    """
    Finds the log entries of a log service matching the given ID

    Args:
        context: The Redfish client object with an open session
        container_type: The type of resource containing the log service (manager, system, or chassis)
        container_id: The container instance with the log service; if None, perform on the only container
        log_service_id: The log service with the logs; if None, perform on the only log service

    Returns:
        An array of log entries
    """

    log_service = get_log_service( context, container_type, container_id, log_service_id )

    # Read in the log entries
    log_entries = []
    log_entry_col = context.get( log_service.dict["Entries"]["@odata.id"] )
    log_entries.extend( log_entry_col.dict["Members"] )

    # If a next link is provided, iterate over it and add to the log entry list
    while "Members@odata.nextLink" in log_entry_col.dict:
        log_entry_col = context.get( log_entry_col.dict["Members@odata.nextLink"] )
        log_entries.extend( log_entry_col.dict["Members"] )

    return log_entries

def print_log_entries( log_entries, details = False ):
    """
    Prints a set of log entries in a table

    Args:
        log_entries: The log entries to print
        details: Flag indicating if details should be displayed
    """

    # Set up templates
    console_size = shutil.get_terminal_size(fallback=(80, 24))
    message_size = console_size.columns - 38
    entry_line_format = "  {:5s} | {:25s} | {}"
    detail_line_format = "  {:33s} | {}: {}"
    detail_list = [ "Severity", "EntryType", "OemRecordFormat", "EntryCode", "OemLogEntryCode", "SensorType", "OemSensorType",
        "GeneratorId", "SensorNumber", "EventType", "EventId", "EventGroupId", "MessageId", "MessageArgs" ]
    print( entry_line_format.format( "Id", "Timestamp", "Message" ) )

    # Go through each entry and print the info
    for entry in log_entries:
        timestamp_property = "Created"
        if "EventTimestamp" in entry:
            timestamp_property = "EventTimestamp"
        print( entry_line_format.format( entry["Id"], entry.get( timestamp_property, "Unknown" ), entry["Message"].replace( "\n", "; " )[:message_size] ) )
        if details:
            for detail in detail_list:
                if detail in entry:
                    print( detail_line_format.format( "", detail, entry[detail] ) )

def clear_log_entries( context, container_type = log_container.MANAGER, container_id = None, log_service_id = None ):
    """
    Clears the log entries of a log service matching the given ID

    Args:
        context: The Redfish client object with an open session
        container_type: The type of resource containing the log service (manager, system, or chassis)
        container_id: The container instance with the log service; if None, perform on the only container
        log_service_id: The log service with the logs; if None, perform on the only log service

    Returns:
        The response of the action
    """

    log_service = get_log_service( context, container_type, container_id, log_service_id )

    # Find the ClearLog action
    if "Actions" not in log_service.dict:
        raise RedfishClearLogNotFoundError( "Log service does not support ClearLog" )
    if "#LogService.ClearLog" not in log_service.dict["Actions"]:
        raise RedfishClearLogNotFoundError( "Log service does not support ClearLog" )
    clear_uri = log_service.dict["Actions"]["#LogService.ClearLog"]["target"]

    # Clear the log
    response = context.post( clear_uri, body = {} )
    verify_response( response )
    return response

def get_log_service( context, container_type = log_container.MANAGER, container_id = None, log_service_id = None ):
    """
    Finds a log service matching the given ID and returns its resource

    Args:
        context: The Redfish client object with an open session
        container_type: The type of resource containing the log service (manager, system, or chassis)
        container_id: The container instance to locate; if None, perform on the only container
        log_service_id: The log service to locate; if None, perform on the only log service

    Returns:
        The log service resource
    """

    # Get the Service Root to find the resource collection
    service_root = context.get( "/redfish/v1/" )
    if container_type.value not in service_root.dict:
        # No resource collection
        raise RedfishLogServiceNotFoundError( "Service does not contain a {} collection".format( container_type.value ) )

    # Get the resource collection and iterate through its collection to find the matching container instance
    avail_resources = []
    resource_col = context.get( service_root.dict[container_type.value]["@odata.id"] )
    if container_id is None:
        if len( resource_col.dict["Members"] ) == 1:
            resource = context.get( resource_col.dict["Members"][0]["@odata.id"] )
            container_id = resource.dict["Id"]
        else:
            for resource_member in resource_col.dict["Members"]:
                resource = context.get( resource_member["@odata.id"] )
                avail_resources.append( resource.dict["Id"] )
            raise RedfishLogServiceNotFoundError( "Service does not contain exactly one resource in {}; a target container needs to be specified: {}".format( container_type.value, ", ".join( avail_resources ) ) )
    else:
        container_found = False
        for resource_member in resource_col.dict["Members"]:
            resource = context.get( resource_member["@odata.id"] )
            avail_resources.append( resource.dict["Id"] )
            if resource.dict["Id"] == container_id:
                container_found = True
                break
        if not container_found:
            raise RedfishLogServiceNotFoundError( "Service does not contain a resource in {} called {}; valid resources: {}".format( container_type.value, container_id, ", ".join( avail_resources ) ) )

    # Go through the container to find the log service collection
    if "LogServices" not in resource.dict:
        # No log services
        raise RedfishLogServiceNotFoundError( "{} does not contain a log service collection".format( container_id ) )

    # Get the log service collection and iterate through its collection
    avail_logs = []
    log_serv_col = context.get( resource.dict["LogServices"]["@odata.id"] )
    if log_service_id is None:
        if len( log_serv_col.dict["Members"] ) == 1:
            return context.get( log_serv_col.dict["Members"][0]["@odata.id"] )
        else:
            for log_serv_member in log_serv_col.dict["Members"]:
                log_serv = context.get( log_serv_member["@odata.id"] )
                avail_logs.append( log_serv.dict["Id"] )
            raise RedfishLogServiceNotFoundError( "{} does not contain exactly one log service; a target log service needs to be specified: {}".format( container_id, ", ".join( avail_logs ) ) )
    else:
        for log_serv_member in log_serv_col.dict["Members"]:
            log_serv = context.get( log_serv_member["@odata.id"] )
            avail_logs.append( log_serv.dict["Id"] )
            if log_serv.dict["Id"] == log_service_id:
                return log_serv

    raise RedfishLogServiceNotFoundError( "{} does not contain a log service called {}; valid log services: {}".format( container_id, log_service_id, ", ".join( avail_logs ) ) )



#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2020 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/master/LICENSE.md

"""
Messages Module

File : messages.py

Brief : This file contains the definitions and functionalities for interacting
        with Messages for a given Redfish service
"""

class RedfishOperationFailedError( Exception ):
    """
    Raised when an operation has failed (HTTP Status >= 400)
    """
    pass

def print_error_payload( response ):
    """
    Prints an error payload, which can also be used for action responses

    Args:
        response: The response to print
    """

    # Check if this response has no body
    if response.dict is None:
        if response.status >= 400:
            print( "Failed" )
        else:
            print( "Success" )
        return

    print( get_error_messages( response ) )

def get_error_messages( response ):
    """
    Builds a string based on the error messages in the payload

    Args:
        response: The response to print

    Returns:
        The string containing error messages
    """

    # Pull out the error payload and the messages
    out_string = response.dict["error"]["message"]
    if "@Message.ExtendedInfo" in response.dict["error"]:
        for message in response.dict["error"]["@Message.ExtendedInfo"]:
            if "Message" in message:
                out_string = out_string + "\n" + message["Message"]
            else:
                out_string = out_string + "\n" + message["MessageId"]

    return out_string

def verify_response( response ):
    """
    Verifies a response and raises an exception if there was a failure

    Args:
        response: The response to verify
    """

    if response.status >= 400:
        exception_string = get_error_messages( response )
        raise RedfishOperationFailedError( "Operation failed: HTTP {}\n{}".format( response.status, exception_string ) )

    return



#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2020 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/master/LICENSE.md

"""
Redfish Logs

File : rf_logs.py

Brief : This script uses the redfish_utilities module to manage logs
"""

import argparse
import redfish
import redfish_utilities

# Get the input arguments
argget = argparse.ArgumentParser( description = "A tool to manage logs on a Redfish service" )
argget.add_argument( "--user", "-u", type = str, required = True, help = "The user name for authentication" )
argget.add_argument( "--password", "-p",  type = str, required = True, help = "The password for authentication" )
argget.add_argument( "--rhost", "-r", type = str, required = True, help = "The address of the Redfish service (with scheme)" )
argget.add_argument( "--manager", "-m", type = str, nargs = "?", default = False, help = "The ID of the manager containing the log service" )
argget.add_argument( "--system", "-s", type = str, nargs = "?", default = False, help = "The ID of the system containing the log service" )
argget.add_argument( "--chassis", "-c", type = str, nargs = "?", default = False, help = "The ID of the chassis containing the log service" )
argget.add_argument( "--log", "-l", type = str, help = "The ID of the resource containing the log service" )
argget.add_argument( "--details", "-details", action = "store_true", help = "Indicates details to be shown for each log entry" )
argget.add_argument( "--clear", "-clear", action = "store_true", help = "Indicates if the log should be cleared" )
args = argget.parse_args()

# Determine the target log service based on the inputs
# Effectively if the user gives multiple targets, some will be ignored
container_type = redfish_utilities.log_container.MANAGER
container_id = None
if args.manager != False:
    container_type = redfish_utilities.log_container.MANAGER
    container_id = args.manager
elif args.system != False:
    container_type = redfish_utilities.log_container.SYSTEM
    container_id = args.system
elif args.chassis != False:
    container_type = redfish_utilities.log_container.CHASSIS
    container_id = args.chassis

# Set up the Redfish object
redfish_obj = redfish.redfish_client( base_url = args.rhost, username = args.user, password = args.password )
redfish_obj.login( auth = "session" )

try:
    # Either clear the logs or get/print the logs
    if args.clear:
        # Clear log was requested
        print( "Clearing the log..." )
        response = redfish_utilities.clear_log_entries( redfish_obj, container_type, container_id, args.log )
        response = redfish_utilities.poll_task_monitor( redfish_obj, response )
        redfish_utilities.verify_response( response )
    else:
        # Print log was requested
        log_entries = redfish_utilities.get_log_entries( redfish_obj, container_type, container_id, args.log )
        redfish_utilities.print_log_entries( log_entries, args.details )
finally:
    # Log out
    redfish_obj.logout()



#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2020 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/master/LICENSE.md

"""
Redfish Boot Override

File : rf_boot_override.py

Brief : This script uses the redfish_utilities module to manage a one time boot
        override of a system
"""

import argparse
import redfish
import redfish_utilities

# Get the input arguments
argget = argparse.ArgumentParser( description = "A tool to perform a one time boot override of a system" )
argget.add_argument( "--user", "-u", type = str, required = True, help = "The user name for authentication" )
argget.add_argument( "--password", "-p",  type = str, required = True, help = "The password for authentication" )
argget.add_argument( "--rhost", "-r", type = str, required = True, help = "The address of the Redfish service (with scheme)" )
argget.add_argument( "--system", "-s", type = str, help = "The ID of the system to set" )
argget.add_argument( "--target", "-t", type = str, help = "The target boot device; if not provided the tool will display the current boot settings" )
argget.add_argument( "--uefi", "-uefi", type = str, help = "If target is 'UefiTarget', the UEFI Device Path of the device to boot.  If target is 'UefiBootNext', the UEFI Boot Option string of the device to boot." )
argget.add_argument( "--reset", "-reset", action = "store_true", help = "Signifies that the system is reset after the boot override is set" )
args = argget.parse_args()

# Set up the Redfish object
redfish_obj = redfish.redfish_client( base_url = args.rhost, username = args.user, password = args.password )
redfish_obj.login( auth = "session" )

try:
    if args.target is None:
        # Target not specified; just get the settings and display them
        boot = redfish_utilities.get_system_boot( redfish_obj, args.system )
        redfish_utilities.print_system_boot( boot )
    else:
        # Build and send the boot request based on the arguments given
        uefi_target = None
        boot_next = None
        boot_mode = "Once"
        if args.target == "UefiTarget":
            uefi_target = args.uefi
        if args.target == "UefiBootNext":
            boot_next = args.uefi
        if args.target == "None":
            print( "Disabling one time boot..." )
            boot_mode = "Disabled"
        else:
            print( "Setting a one time boot for {}...".format( args.target ) )
        redfish_utilities.set_system_boot( redfish_obj, args.system, args.target, boot_mode, None, uefi_target, boot_next )

        # Reset the system if requested
        if args.reset:
            print( "Resetting the system..." )
            response = redfish_utilities.system_reset( redfish_obj, args.system )
            response = redfish_utilities.poll_task_monitor( redfish_obj, response )
            redfish_utilities.verify_response( response )
finally:
    # Log out
    redfish_obj.logout()



#! /usr/bin/python
# Copyright Notice:
# Copyright 2019-2020 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Tacklebox/blob/master/LICENSE.md

"""
Accounts Module

File : accounts.py

Brief : This file contains the definitions and functionalities for managing
        accounts on a Redfish Service
"""

from .messages import verify_response

class RedfishAccountCollectionNotFoundError( Exception ):
    """
    Raised when the Account Service or Account Collection cannot be found
    """
    pass

class RedfishAccountNotFoundError( Exception ):
    """
    Raised when the Account Service or Account Collection cannot be found
    """
    pass

class RedfishAccountNotAddedError( Exception ):
    """
    Raised when the requested account was not added
    """
    pass

def get_users( context ):
    """
    Collects user information from a Redfish service

    Args:
        context: The Redfish client object with an open session

    Returns:
        A list containing all users
    """

    user_list = []

    # Go through each Account in the Account Collection
    account_col = context.get( get_account_collection( context ), None )
    for account_member in account_col.dict["Members"]:
        account = context.get( account_member["@odata.id"], None )
        account_info = {
            "UserName": account.dict["UserName"],
            "RoleId": account.dict["RoleId"],
            "Locked": account.dict.get( "Locked", False ),
            "Enabled": account.dict.get( "Enabled", True )
        }

        # Some implementations always expose "slots" for users; ignore empty slots
        if account_info["UserName"] == "" and account_info["Enabled"] == False:
            continue

        user_list.append( account_info )

    return user_list

def print_users( user_list ):
    """
    Prints the user list into a table

    Args:
        user_list: The user list to print
    """

    user_line_format = "  {:20s} | {:20s} | {:10s} | {:10s}"
    print( "" )
    print( user_line_format.format( "Name", "Role", "Locked", "Enabled" ) )
    for user in user_list:
        print( user_line_format.format( user["UserName"], user["RoleId"], str( user["Locked"] ), str( user["Enabled"] ) ) )
    print( "" )

def add_user( context, user_name, password, role ):
    """
    Adds a new user account

    Args:
        context: The Redfish client object with an open session
        user_name: The name of the user to add
        password: The password for the new user
        role: The role for the new user

    Returns:
        The response of the POST
    """

    # Create the new user
    payload = {
        "UserName": user_name,
        "Password": password,
        "RoleId": role
    }
    account_col_uri = get_account_collection( context )
    response = context.post( account_col_uri, body = payload )
    if response.status == 405:
        # Some implementations allocate slots for users and don't allow adding in the proper sense
        # Find an empty slot to use
        account_added = False
        account_col = context.get( account_col_uri )
        for account_member in account_col.dict["Members"]:
            account = context.get( account_member["@odata.id"] )
            if account.dict["UserName"] == "" and not account.dict.get( "Enabled", True ):
                # Empty slot found; PATCH it
                response = context.patch( account_member["@odata.id"], body = payload, headers = { "If-Match": account.getheader( "ETag" ) } )
                if response.status < 400:
                    # These implementations also might restrict which slots to modify...
                    account_added = True
                    break
        if not account_added:
            raise RedfishAccountNotAddedError( "Failed to add user '{}'; user may already exist or there is not enough space for the new user".format( user_name ) )
    verify_response( response )
    return response

def delete_user( context, user_name ):
    """
    Deletes an existing user account

    Args:
        context: The Redfish client object with an open session
        user_name: The name of the user to delete

    Returns:
        The response of the DELETE
    """

    # Find the user to delete
    user_uri, user_info = get_user( context, user_name )

    # Delete the user
    response = context.delete( user_uri )
    if response.status == 405:
        # Some implementations keep slots around and don't allow for deleting in the proper sense
        # Some also do not allow for both Enabled and UserName to be modified simultaneously
        modify_user( context, user_name, new_enabled = False )
        return modify_user( context, user_name, new_name = "" )
    verify_response( response )
    return response

def modify_user( context, user_name, new_name = None, new_password = None, new_role = None, new_locked = None, new_enabled = None ):
    """
    Modifies an existing user account

    Args:
        context: The Redfish client object with an open session
        user_name: The name of the user to modify
        new_name: The new name of the user
        new_password: The new password of the user
        new_role: The new role of the user
        new_locked: The new locked flag of the user
        new_enabled: The new enabled flag of the user

    Returns:
        The response of the PATCH
    """

    # Get the current user info
    user_uri, user_info = get_user( context, user_name )

    # Build the payload for the new user
    new_info = {}
    if new_name is not None:
        new_info["UserName"] = new_name
    if new_password is not None:
        new_info["Password"] = new_password
    if new_role is not None:
        new_info["RoleId"] = new_role
    if new_locked is not None:
        new_info["Locked"] = new_locked
    if new_enabled is not None:
        new_info["Enabled"] = new_enabled

    # Update the user
    response = context.patch( user_uri, body = new_info, headers = { "If-Match": user_info.getheader( "ETag" ) } )
    verify_response( response )
    return response

def get_account_collection( context ):
    """
    Finds the account collection for the Redfish service

    Args:
        context: The Redfish client object with an open session

    Returns:
        The URI for the account collection
    """

    # Get the Service Root to find the Account Service
    service_root = context.get( "/redfish/v1/" )
    if "AccountService" not in service_root.dict:
        # No Account Service
        raise RedfishAccountCollectionNotFoundError( "Service does not contain an Account Service" )

    # Get the Account Service to find the Account Collection
    account_service = context.get( service_root.dict["AccountService"]["@odata.id"] )
    if "Accounts" not in account_service.dict:
        # No Account Collection
        raise RedfishAccountCollectionNotFoundError( "Service does not contain an Account Collection" )

    return account_service.dict["Accounts"]["@odata.id"]

def get_user( context, user_name ):
    """
    Finds a user within the Redfish service

    Args:
        context: The Redfish client object with an open session
        user_name: The name of the user to find

    Returns:
        The URI for the user account
        The contents of the user account resource
    """

    avail_users = []
    account_col = context.get( get_account_collection( context ) )
    for account_member in account_col.dict["Members"]:
        account = context.get( account_member["@odata.id"] )

        # Some implementations always expose "slots" for users; ignore empty slots
        if account.dict["UserName"] == "" and not account.dict.get( "Enabled", True ):
            continue

        avail_users.append( account.dict["UserName"] )

        # Check if the name matches
        if account.dict["UserName"] == user_name:
            return account_member["@odata.id"], account

    # No matches found
    raise RedfishAccountCollectionNotFoundError( "User '{}' is not found; valid users: {}".format( user_name, ", ".join( avail_users ) ) )
